using System;
using System.Collections.Generic;
using System.Diagnostics;
using Apache.Cassandra;
using Cassandraemon.Builder;


namespace Cassandraemon.Operators
{


	class GetCountCassandraOperator<T> : CassandraOperatorBase<T>
	{
		public byte[] Key { get; set; }
		
		public ColumnParent ColumnParent { get; set; }
		
		public SlicePredicate Predicate { get; set; }
	
		public ConsistencyLevel ConsistencyLevel { get; set; }
	
	
		public GetCountCassandraOperator () { }
		
		
		protected override object OperateInner ()
		{
            var ctx = this.Context;
            var sw = Stopwatch.StartNew();
            int count;
            try {
                count = Client.get_count(Key, ColumnParent, Predicate, ConsistencyLevel);
            } catch (Exception err) {
                ctx.OnExecuted(CassandraExecuteEventArgs.FromOperation(this, err, sw));
                throw;
            }
            ctx.OnExecuted(CassandraExecuteEventArgs.FromOperation(this, count, sw));

			T t = (T)Activator.CreateInstance(typeof(T));
			
			if(t is short ||
			   t is int ||
			   t is long ||
			   t is ushort ||
			   t is uint ||
			   t is ulong)
			{
				return count;
			}
			else
			{
				throw new InvalidOperationException("You specify not numeric type to CassandraEntity's GenericType in get_count operation."); 
			}
		}

#if TRACE
        protected internal override void GetTraceData(List<object> data)
        {
            base.GetTraceData(data);
            data.Add("get_count");
            var cp = this.ColumnParent;
            data.Add(cp.Column_family);
        }
#endif

		public static ICassandraOperator<T> Create(CassandraOperatorCondition<T> condition)
		{
			if(!condition.IsSetKey) throw new InvalidOperationException("You must specify Key in case of get_count operation.");
			if(condition.IsSetKeys) throw new InvalidOperationException("You can't specify Keys in case of get_count operation.");
			
			if(condition.EqIndexExpression.Count  > 0 ||
			   condition.GteIndexExpression.Count > 0 ||
			   condition.GtIndexExpression.Count  > 0 ||
			   condition.LteIndexExpression.Count > 0 ||
			   condition.LtIndexExpression.Count  > 0)
				throw new InvalidOperationException("You can't specify index in case of get_count operation.");
				
			if(condition.IsSetStartKey || 
			   condition.IsSetFinishKey || 
			   condition.IsSetStartToken || 
			   condition.IsSetFinishToken ||
			   condition.IsSetKeyCount)
				throw new InvalidOperationException("You can't specify range key in case of get_count operation.");
			
			if(condition.IsSetColumns &&
			  (condition.IsSetStartColumn || 
			   condition.IsSetFinishColumn || 
			   condition.IsSetReverse ||
			   condition.IsSetCount))
				throw new InvalidOperationException("You can't specify both Columns and range column in case of get_count operation.");
			
				
			var oper = new GetCountCassandraOperator<T>();
			
			oper.Key = condition.Key;
			oper.ConsistencyLevel = condition.ConsistencyLevel;
			oper.PostProcessor = condition.PostProcessor;
			
			oper.ColumnParent = new ColumnParent();
			oper.ColumnParent.Column_family = condition.ColumnFamily;
			oper.ColumnParent.Super_column = condition.SuperColumn;
			
			oper.Predicate = new SlicePredicate();
			oper.Predicate.Column_names = condition.Columns;
			if(condition.Columns == null)
			{
				oper.Predicate.Slice_range = new SliceRange();
				oper.Predicate.Slice_range.Start = condition.StartColumn;
				oper.Predicate.Slice_range.Finish = condition.FinishColumn;
				oper.Predicate.Slice_range.Reversed = condition.Reverse;
				oper.Predicate.Slice_range.Count = condition.Count;
			}	
			
			Type type = typeof(T);
			
			if(type.Equals(typeof(short)) ||
			   type.Equals(typeof(int)) ||
			   type.Equals(typeof(long)) ||
			   type.Equals(typeof(ushort)) ||
			   type.Equals(typeof(uint)) ||
			   type.Equals(typeof(ulong)))
		   	{
		   		// Do Nothing
			}
			else
			{
				throw new InvalidOperationException("Return type specified not numeric type in get_count operation."); 
			}
			
			return oper;
		}

	}
}
