using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Apache.Cassandra;
using Cassandraemon.Builder;


namespace Cassandraemon.Operators
{


	class MultigetSliceCassandraOperator<T> : CassandraOperatorBase<T>
	{
		public List<byte[]> Keys { get; set; }
		
		public ColumnParent ColumnParent { get; set; }
		
		public SlicePredicate Predicate { get; set; }
		
		public ConsistencyLevel ConsistencyLevel { get; set; }
	
	
		public MultigetSliceCassandraOperator () { }
		
		
		protected override object OperateInner ()
		{
            var ctx = this.Context;
            var sw = Stopwatch.StartNew();
            Dictionary<byte[], List<ColumnOrSuperColumn>> columnListMap;
            try {
                columnListMap = Client.multiget_slice(Keys, ColumnParent, Predicate, ConsistencyLevel);
            } catch (Exception err) {
                ctx.OnExecuted(CassandraExecuteEventArgs.FromOperation(this, err, sw));
                throw;
            }
            ctx.OnExecuted(CassandraExecuteEventArgs.FromOperation(this, columnListMap, sw));

			var entityList = new List<CassandraEntity<T>>();
			
			foreach(var kv in columnListMap)
			{
				entityList.Add(MakeCassandraEntity(kv.Value, kv.Key, this.ColumnParent));
			}
			
			return entityList;
		}

#if TRACE
        protected internal override void GetTraceData(List<object> data)
        {
            base.GetTraceData(data);
            data.Add("multiget_slice");
            var cp = this.ColumnParent;
            data.Add(cp.Column_family);
        }
#endif

		public static ICassandraOperator<T> Create(CassandraOperatorCondition<T> condition)
		{
			if(condition.IsSetKey) throw new InvalidOperationException("You can't specify Key in case of multiget_slice operation.");
			if(!condition.IsSetKeys) throw new InvalidOperationException("You must specify Keys in case of multiget_slice operation.");
			
			if(condition.IsSetStartKey || 
			   condition.IsSetFinishKey || 
			   condition.IsSetStartToken || 
			   condition.IsSetFinishToken ||
			   condition.IsSetKeyCount)
				throw new InvalidOperationException("You can't specify range key in case of multiget_slice operation.");
	
			if(condition.IsSetColumns &&
			  (condition.IsSetStartColumn || 
			   condition.IsSetFinishColumn || 
			   condition.IsSetReverse ||
			   condition.IsSetCount))
				throw new InvalidOperationException("You can't specify both Columns and range column in case of multiget_slice operation.");
			
			
			var oper = new MultigetSliceCassandraOperator<T>();
			
			oper.Keys = condition.Keys;
			oper.ConsistencyLevel = condition.ConsistencyLevel;
			oper.PostProcessor = condition.PostProcessor;
			
			oper.ColumnParent = new ColumnParent();
			oper.ColumnParent.Column_family = condition.ColumnFamily;
			oper.ColumnParent.Super_column = condition.SuperColumn;
			
			oper.Predicate = new SlicePredicate();
			oper.Predicate.Column_names = condition.Columns;
			if(condition.Columns == null)
			{
				oper.Predicate.Slice_range = new SliceRange();
				oper.Predicate.Slice_range.Start = condition.StartColumn;
				oper.Predicate.Slice_range.Finish = condition.FinishColumn;
				oper.Predicate.Slice_range.Reversed = condition.Reverse;
				oper.Predicate.Slice_range.Count = condition.Count;
			}
			
			
			Type type = typeof(T);
			
			if(type.Equals(typeof(Column)) || type.Equals(typeof(CounterColumn)))
		   	{
		   		// Set No Limit because of enable FirstColumn method etc.
			}
			else if(type.Equals(typeof(SuperColumn)) || type.Equals(typeof(CounterSuperColumn)))
			{
				if(condition.IsSetSuperColumn) throw new InvalidOperationException("You can't specify SuperColumn in case of multiget_slice SuperColumn.");
			}
			else if(type.Equals(typeof(List<Column>)) || type.Equals(typeof(List<CounterColumn>)))
			{
				// Do Nothing
			}
			else if(type.Equals(typeof(List<SuperColumn>)) || type.Equals(typeof(List<CounterSuperColumn>)))
			{
				if(condition.IsSetSuperColumn) throw new InvalidOperationException("You can't specify SuperColumn in case of multiget_slice SuperColumn.");
			}
			else
			{
				throw new InvalidOperationException("You must specify Column or SuperColumn or CounterColumn or CounterSuperColumn or List<Column> or List<SuperColumn> or List<CounterColumn> or List<CounterSuperColumn> to CassandraEntity's GenericType in multiget_slice operation."); 
			}
			
			return oper;
		}

	}
}
